open Format
open Lexing

exception Syntax_error of position * string

let print ppf (lb, p, msg) =
  let pos0 = -lb.lex_abs_pos in
  let end_pos = lb.lex_buffer_len - pos0 - 1 in
  let b = Buffer.create 80 in
  let rec loop pos =
    Buffer.add_char b lb.lex_buffer.[pos + pos0];
    if pos = end_pos then 
      Buffer.contents b
    else
      if lb.lex_buffer.[pos + pos0] = '\n' then 
        if pos < p.pos_cnum then
          (Buffer.clear b; loop (pos + 1))
        else if pos > p.pos_cnum then
          Buffer.contents b
        else
          loop (pos + 1)
      else 
        loop (pos + 1)
  in
  let s = if pos0 < 0 then "" else loop 0 in
  let column = p.pos_cnum - p.pos_bol in
  fprintf ppf "File %S at line %d, character %d:@.%s%s^@.%s@." 
    p.pos_fname p.pos_lnum column
    s (String.make column ' ') msg
